import falcon

class Authentication:
    def __init__(self, auth):
        self.auth = auth
        
    def process_request(self, req, resp):
        bearer_auth = req.auth

        challenges = ['Token type="Bearer"']

        if bearer_auth is None:
            description = "Please provide an auth token as part of the request."

            raise falcon.HTTPUnauthorized(
                "Auth token required", description, challenges
            )

        try:
            id_token = bearer_auth.split()[1]
            check_token = auth.verify_id_token(id_token, check_revoked=True)

        except:
            description = (
                "The provided auth token has been revoked. "
                "Please request a new token and try again."
            )

            raise falcon.HTTPUnauthorized(
                "Authentication required", description, challenges
            )

        try:
            decoded_token = auth.verify_id_token(id_token)
            req.context.token = decoded_token
            return

        except:
            description = (
                "The provided auth token is not valid. "
                "Please request a new token and try again."
            )

            raise falcon.HTTPUnauthorized(
                "Authentication required", description, challenges
            )


from firebase_admin import firestore

def create_collection_query(collection, params=None):
    if params is None:
        return collection
    # if req.params:
    #     params = self._validate_q(req.params)

    #     if params.order_by:
    #         query = query.order_by(
    #             field_path=params.order_by, direction=params.order_dir
    #         )

    return collection


def parse_docs(self, docs):
    doc_list = []
    for doc in docs:

        doc_list.append(
            {
            "id": doc.id,
            "type": self._resource_type,
            "attributes": self._validate_get(doc.to_dict()).dict(),
            }
        )
    return doc_list


# def one_to_many_post(db):
#     transaction = db.transaction()
    
#     @firestore.transactional
    


from pydantic import BaseModel, validator


class BaseSchema(BaseModel):
    class Config:
        extra = "allow"


class BaseQuerySchema(BaseModel):
    where: list = None
    limit: int = 100
    offset: int = 0
    order_by: str = "ASCENDING"

    @validator("where", pre=True)
    def split_str(cls, v):
        if v is None:
            return v
        return v.split(" AND ")



import json

import falcon
from firebase_admin.exceptions import AlreadyExistsError, FirebaseError, NotFoundError
from pydantic import ValidationError

from ._schemas import BaseQuerySchema, BaseSchema
from ._utils import create_collection_query, parse_docs
from .authorize import Authorize


class BaseResource:
    # pylint: disable=no-member
    base_schema = BaseSchema
    base_query_schema = BaseQuerySchema
    authorize = Authorize
    valid_error = ValidationError

    def __init__(
        self,
        db,
        resource_type: str,
        id_alias="id",
        db_path: str = None,
        schema_post=None,
        schema_get=None,
        schema_put=None,
        schema_patch=None,
        schema_q=None,
        allowed: list = None,
    ):
        self._db = db
        self._db_path = db_path
        self._id_alias = id_alias
        self._resource_type = resource_type
        self._schema_post = schema_post
        self._schema_get = schema_get
        self._schema_put = schema_put
        self._schema_patch = schema_patch
        self._schema_q = schema_q
        self.allowed = allowed or []

    def get_resource_link(self, req):
        if req.forwarded_uri is None:
            return req.uri
        return req.forwarded_uri

    def get_db_path(self, path):
        if self._db_path is None:
            return path[1:]
        else:
            return f"{self._db_path}"

    def get_doc_path(self, path, params):
        if self._db_path is None:
            return path[1:]
        else:
            return f"{self._db_path}/{params.get(self._id_alias)}"

    def _validate(self, schema, data):
        try:
            if schema is None:
                return self.base_schema.parse_obj(data)
            return schema.parse_obj(data)

        except self.valid_error as e:
            title = "ValidationError"
            description = str(e)
            raise falcon.HTTPBadRequest(title=title, description=description)

    def _validate_post(self, data):
        return self._validate(self._schema_post, data)

    def _validate_get(self, data):
        return self._validate(self._schema_get, data)

    def _validate_put(self, data):
        return self._validate(self._schema_put, data)

    def _validate_patch(self, data):
        return self._validate(self._schema_patch, data)

    def _validate_q(self, data):
        if self._schema_q is None:
            return self._validate(self.base_query_schema, data)
        return self._validate(self._schema_q, data)

    @falcon.before(authorize("on_get"))
    def on_get(self, req, resp, **kwargs):
        try:
            db_path = self.get_db_path(req.path)
            col_ref = self._db.collection(db_path)

            params = self._validate_q(req.params)
            query_ref = create_collection_query(col_ref, params)

            docs = query_ref.get()
            validated = [
                {
                    "id": doc.id,
                    "type": self._resource_type,
                    "attributes": self._validate_get(doc.to_dict()).dict(),
                }
                for doc in docs
            ]

            resp.status = falcon.HTTP_200
            resp.media = {"links": self.get_resource_link(req), "data": validated}

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)

    @falcon.before(authorize("on_get_doc"))
    def on_get_doc(self, req, resp, **kwargs):
        try:
            db_path = self.get_doc_path(req.path, kwargs)
            doc_ref = self._db.document(db_path)

            doc = doc_ref.get()
            validated = self._validate_get(doc.to_dict())

            resp.status = falcon.HTTP_200
            resp.append_header("Location", self.get_resource_link(req))
            resp.media = {
                "links": {
                    "self": self.get_resource_link(req),
                    "data": {"id": doc_ref.id, "attributes": validated.dict()},
                }
            }

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)

    @falcon.before(authorize("on_post"))
    def on_post(self, req, resp, **kwargs):
        try:
            validated = self._validate_post(req.media)
            valid_dict = validated.dict()

            db_path = self.get_db_path(req.path)
            col_ref = self._db.collection(db_path)

            if hasattr(validated, "id"):
                doc_ref = col_ref.document(validated.id)
                del valid_dict["id"]
                doc_ref.create(valid_dict)

            else:
                doc_ref = col_ref.document()
                doc_ref.create(valid_dict)

            resp.status = falcon.HTTP_201
            resp.append_header("Location", self.get_resource_link(req))
            resp.media = {
                "data": {
                    "id": doc_ref.id,
                    "type": self._resource_type,
                    "attributes": valid_dict,
                },
                "links": {"self": self.get_resource_link(req)},
            }

        except AlreadyExistsError as e:
            raise falcon.HTTPConflict(code=e.code, description=e.message)

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)

    @falcon.before(authorize("on_put_doc"))
    def on_put_doc(self, req, resp, **kwargs):
        try:
            validated = self._validate_put(req.media)

            db_path = self.get_doc_path(req.path, kwargs)
            doc_ref = self._db.document(db_path)

            doc_ref.set(validated.dict())

            resp.status = falcon.HTTP_200
            resp.append_header("Location", self.get_resource_link(req))
            resp.media = {
                "data": {
                    "id": doc_ref.id,
                    "type": self._resource_type,
                    "attributes": validated.dict(),
                },
                "links": {"self": self.get_resource_link(req)},
            }

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)

    @falcon.before(authorize("on_put_doc"))
    def on_patch_doc(self, req, resp, **kwargs):
        try:
            validated = self._validate_patch(req.media)

            db_path = self.get_doc_path(req.path, kwargs)
            doc_ref = self._db.document(db_path)

            doc_ref.update(validated.dict())

            resp.status = falcon.HTTP_200
            resp.append_header("Location", self.get_resource_link(req))
            resp.media = {
                "data": {
                    "id": doc_ref.id,
                    "type": self._resource_type,
                    "attributes": validated.dict(),
                },
                "links": {"self": self.get_resource_link(req)},
            }

        except NotFoundError as e:
            raise falcon.HTTPNotFound(code=e.code, description=e.message)

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)

    @falcon.before(authorize("on_delete_doc"))
    def on_delete_doc(self, req, resp, **kwargs):
        try:
            db_path = self.get_doc_path(req.path, kwargs)
            doc_ref = self._db.document(db_path)

            doc_ref.delete()

        except NotFoundError as e:
            raise falcon.HTTPNotFound(code=e.code, description=e.message)

        except FirebaseError as e:
            raise falcon.HTTPBadRequest(code=e.code, description=e.message)



from firefalcon.baseresource import BaseResource


class OneToManyResource(BaseResource):
    def __init__(
        self,
        db,
        parent_path: str,
        parent_field_path: str,
        is_subcollection: bool = False,
        **kwargs
    ):
        super().__init__(**kwargs)

    def on_get(self, req, resp):
        pass

    def on_get_doc(self, req, resp, doc):
        pass

    def on_post(self, req, resp):
        try:
            validated = self._validate_post(req.media)
            valid_dict = validated.dict()
            
            

        except:
            pass

    def on_put_doc(self, req, resp, doc):
        pass

    def on_delete_doc(self, req, resp, doc):
        pass



import falcon


class Authorize(object):
    def __init__(self, responder=None):
        self._responder = responder

    def __call__(self, req, resp, resource, params):
        print(resource.allowed)       
        if not req.context.get("token"):
            return
        
        if not req.context.token.get("role") == 'admin':
            if self._responder not in resource.allowed:
                msg = "User not authorized."
                raise falcon.HTTPUnauthorized("Unauthorized", msg)
        return



import importlib.util
import os.path
from setuptools import setup, find_packages


HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, "README.md")) as readme_file:
    README = readme_file.read()


DESCRIPTION = "Tools for integrating Falcon applications with Firebase."

REQUIRES = ["falcon >= 2.0.0", "firebase_admin >= 4.0.0", "pydantic >= 1.0.0"]


setup(
    name="firefalcon",
    version="0.0.4",
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/zqwalker/firefalcon",
    author="Zackary Walker",
    author_email="quinn@soarbi.com",
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Topic :: Database :: Front-Ends",
        "Topic :: Software Development :: Libraries",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="falcon middleware firebase firestore authentication",
    packages=['firefalcon'],
    python_requires=">=3.8",
    install_requires=REQUIRES,
    include_package_data=True,
)